/*
 * SPDX-FileCopyrightText: 2022 Institute Rhea <eduard.filipas@rhea.si>
 *
 * SPDX-License-Identifier: EUPL-1.2
 */

#include "File.h"
#include "algorithm"
#include <sys/stat.h>
#include "filesystem"
#include "iostream"
#include <fstream>


rhea::File::File() {
    name = getUUID() + ".data";
}

rhea::File::~File() {

}

void rhea::File::setType(std::string val) {
    type = val;
}

bool rhea::File::copyFrom(std::string origin) {
	std::filesystem::file_status fs = std::filesystem::status(origin);
	if(!std::filesystem::is_regular_file(fs)){ return false; }

	if (getName().empty()){ return false; }
	return std::filesystem::copy_file(origin, getFullName());
}

std::string rhea::File::getFullName() {
	std::string tmp = "";
	if (path.starts_with(pathDelimiter)){
		tmp = path+name;
    } else{
    	tmp = base+path+name;
    }
	return tmp;
}

void rhea::File::setFullName(std::string file) {
	if(file.empty()) {return;}

	setName(file.substr(file.find_last_of(pathDelimiter)+1));
	file = file.replace(file.end()-getName().length(),file.end(),"");

	if (file.starts_with(base)){
		file = file.replace(file.begin(),file.begin()+getBase().length(),"");
		setPath(file,false);
	}else{
		setPath(file.substr(0,file.find_last_of(pathDelimiter)),false);
	}
}

std::string &rhea::File::getBase() {
    return base;
}

void rhea::File::setBase( std::string val, bool create) {
    base = replaceTilda(val);
    base += base.ends_with(pathDelimiter) ? "" :  pathDelimiter;
    if (create && !dirExists(base)){
        createPath(base);
    }
}

std::string &rhea::File::getPath() {
    return path;
}

void rhea::File::setPath(std::string val, bool create) {
    if (val.empty()) {
    	path = "";
    	return;
    }

	path = replaceTilda(val);
    path += path.ends_with(pathDelimiter) ? "" : pathDelimiter;

    std::string myPath = path;
    if (!path.starts_with("/") && !path.starts_with(base)){
    	myPath = getBase()+path;
    }
    if (create && !dirExists(myPath)){
        createPath(myPath);
    }
}

std::string &rhea::File::getName() {
    return name;
}

void rhea::File::setName(std::string val) {
    name = val;
    std::reverse(val.begin(), val.end());
    val = val.substr(0,val.find('.'));
    std::reverse(val.begin(), val.end());
    setType(val);
}

std::string &rhea::File::getType() {
    return type;
}

void rhea::File::clearObject() {
    name = "";
    path = "";
    type = "";
    length = 0;

}

bool rhea::File::createPath(std::string &path) {
    path = path.ends_with(pathDelimiter) ? path : path + pathDelimiter;
    path = replaceTilda(path);
    std::size_t found = path.find(pathDelimiter);
    bool isOk = true;
    std::string dir = "";
    while(found != path.npos){
        if (found>0) {
            dir = path.substr(0, found );
            if (!dir.empty() && !dirExists(dir)) {
                try {
                    std::filesystem::create_directory(dir);
                } catch (int e) {
                    isOk = false;
                }
            }
        }
        found = path.find(pathDelimiter,found +1);
    }

    return isOk;
}

bool rhea::File::dirExists(std::string &path) {

    struct stat sb;
    path = replaceTilda(path);
    return stat(path.c_str(), &sb) == 0 && S_ISDIR(sb.st_mode);
}

std::string rhea::File::replaceTilda(std::string &val) {
    std::string home_dir = getenv("HOME");

    home_dir = home_dir.ends_with(pathDelimiter) ? home_dir.substr(0,home_dir.length()-1) : home_dir;
    if (!val.empty() && val.starts_with("~")) {
        val = val.replace(0, 1, home_dir);
    }
    return val;
}


bool rhea::File::getFileContent(std::string &out) {
	out = "";
	std::ifstream in(getFullName().c_str());

     // Check if object is valid
     if (!in) {
         std::cerr << "Cannot open the File : " << getFullName() << std::endl;
         return false;
     }
     std::string str;
     // Read the next line from File untill it reaches the end.
     while (std::getline(in, str)) {
         // Line contains string of length > 0 then save it in vector
         if (str.size() > 0) {
        	 out += str;
         }
     }
     //Close The File
     in.close();
     return true;
}

bool rhea::File::zipper_add_file(zipFile zfile, const char *filename, const char* destName)
{
    FILE          *f;
    unsigned char  buf[BUF_SIZE];
    int            ret;
    size_t         red;
    size_t         flen;

    if (zfile == NULL || filename == NULL)
        return false;

    f = fopen(filename, "r");
    if (f == NULL)
        return false;

    fseek(f, 0, SEEK_END);
    flen = ftell(f);
    rewind(f);

    ret = zipOpenNewFileInZip64(zfile, destName, NULL, NULL, 0, NULL, 0, NULL,
            Z_DEFLATED, Z_DEFAULT_COMPRESSION, (flen > 0xffffffff)?1:0);
    if (ret != ZIP_OK) {
        fclose(f);
        return false;
    }

    while ((red = fread(buf, sizeof(*buf), sizeof(buf), f)) > 0) {
        ret = zipWriteInFileInZip(zfile, buf, red);
        if (ret != ZIP_OK) {
            fclose(f);
            zipCloseFileInZip(zfile);
            return false;
        }
    }

    zipCloseFileInZip(zfile);
    return true;
}

bool rhea::File::zipper_add_buf(zipFile zfile, const char *zfilename, const unsigned char *buf, size_t buflen)
{
    int ret;

    if (zfile == NULL || buf == NULL || buflen == 0)
        return false;

    ret = zipOpenNewFileInZip64(zfile, zfilename, NULL, NULL, 0, NULL, 0, NULL,
            Z_DEFLATED, Z_DEFAULT_COMPRESSION, (buflen > 0xffffffff)?1:0);
    if (ret != ZIP_OK)
        return false;

    ret = zipWriteInFileInZip(zfile, buf, buflen);
    zipCloseFileInZip(zfile);
    return ret==ZIP_OK?true:false;
}

bool rhea::File::zipper_add_dir(zipFile zfile, const char *dirname)
{
    char   *temp;
    size_t  len;
    int     ret;

    if (zfile == NULL || dirname == NULL || *dirname == '\0')
        return false;

    len  = strlen(dirname);
    temp = (char*)calloc(1, len+2);
    memcpy(temp, dirname, len+2);
    if (temp[len-1] != '/') {
        temp[len] = '/';
        temp[len+1] = '\0';
    } else {
        temp[len] = '\0';
    }

    ret = zipOpenNewFileInZip64(zfile, temp, NULL, NULL, 0, NULL, 0, NULL, 0, 0, 0);
    if (ret != ZIP_OK)
        return false;
    free(temp);
    zipCloseFileInZip(zfile);
    return ret==ZIP_OK?true:false;
}

rhea::zipper_result_t rhea::File::zipper_read(unzFile zfile, zipper_read_cb_t cb, void *thunk)
{
    unsigned char tbuf[BUF_SIZE];
    int           red;
    int           ret;

    if (zfile == NULL || cb == NULL)
        return ZIPPER_RESULT_ERROR;

    ret = unzOpenCurrentFile(zfile);
    if (ret != UNZ_OK)
        return ZIPPER_RESULT_ERROR;

    while ((red = unzReadCurrentFile(zfile, tbuf, sizeof(tbuf))) > 0) {
        cb(tbuf, red, thunk);
    }

    if (red < 0) {
        unzCloseCurrentFile(zfile);
        return ZIPPER_RESULT_ERROR;
    }

    unzCloseCurrentFile(zfile);
    if (unzGoToNextFile(zfile) != UNZ_OK)
        return ZIPPER_RESULT_SUCCESS_EOF;
    return ZIPPER_RESULT_SUCCESS;
}

void rhea::File::zipper_read_buf_cb(const unsigned char *buf, size_t buflen, void *thunk)
{
    str_builder_t *sb = (str_builder_t*)thunk;
    str_builder_add_str(sb, (const char *)buf, buflen);
}

rhea::zipper_result_t rhea::File::zipper_read_buf(unzFile zfile, unsigned char **buf, size_t *buflen)
{
    str_builder_t   *sb;
    zipper_result_t  ret;

    sb = str_builder_create();
    ret = zipper_read(zfile, zipper_read_buf_cb, sb);
    if (ret != ZIPPER_RESULT_ERROR)
        *buf = (unsigned char *)str_builder_dump(sb, buflen);
    str_builder_destroy(sb);
    return ret;
}

char *rhea::File::zipper_filename(unzFile zfile, bool *isutf8)
{
    char            name[MAX_NAMELEN];
    unz_file_info64 finfo;
    int             ret;

    if (zfile == NULL)
        return NULL;

    ret = unzGetCurrentFileInfo64(zfile, &finfo, name, sizeof(name), NULL, 0, NULL, 0);
    if (ret != UNZ_OK)
        return NULL;
    if (isutf8 != NULL)
        *isutf8 = (finfo.flag & (1<<11))?true:false;
    return strdup(name);
}

bool rhea::File::zipper_isdir(unzFile zfile)
{
    char            name[MAX_NAMELEN];
    unz_file_info64 finfo;
    size_t          len;
    int             ret;

    if (zfile == NULL)
        return false;

    ret = unzGetCurrentFileInfo64(zfile, &finfo, name, sizeof(name), NULL, 0, NULL, 0);
    if (ret != UNZ_OK)
        return false;

    len = strlen(name);
    if (finfo.uncompressed_size == 0 && len > 0 && name[len-1] == '/')
        return true;
    return false;
}

bool rhea::File::zipper_skip_file(unzFile zfile)
{
    if (unzGoToNextFile(zfile) != UNZ_OK)
        return false;
    return true;
}

uint64_t rhea::File::zipper_filesize(unzFile zfile)
{
    unz_file_info64 finfo;
    int             ret;

    if (zfile == NULL)
        return 0;

    ret = unzGetCurrentFileInfo64(zfile, &finfo, NULL, 0, NULL, 0, NULL, 0);
    if (ret != UNZ_OK)
        return 0;
    return finfo.uncompressed_size;
}


bool rhea::File::compresZip(std::string &out) {
    zipFile zfile;
    static const char *zipname = "test.zip";
    static const char *f1_name = "Moneypenny.txt";
    static const char *f2_name = "Bond.txt";
    static const char *f3_name = "Up/M.txt";
    static const char *f4_name = "Down/Q.tt";
    static const char *f1_data = "secretary\n";
    static const char *f2_data = "secret agent\n";
    static const char *f3_data = "top guy\n";
    static const char *f4_data = "bottom guy\n";
    static const char *d1_name = "Up/";
    static const char *d2_name = "Around/";
    static const char *d3_name = "Bound/A/B";

    zfile = zipOpen64(zipname, 0);
    if (zfile == NULL) {
        printf("Could not open %s for zipping\n", zipname);
        zipClose(zfile, NULL);
        return false;
    }

    printf("adding dir: %s\n", d2_name);
    if (!zipper_add_dir(zfile, d2_name)) {
        printf("failed to write dir %s\n", d2_name);
        zipClose(zfile, NULL);
        return false;
    }

    printf("adding dir: %s\n", d3_name);
    if (!zipper_add_dir(zfile, d3_name)) {
        printf("failed to write dir %s\n", d3_name);
        zipClose(zfile, NULL);
        return false;
    }

    printf("adding file: %s\n", f1_name);
    if (!zipper_add_buf(zfile, f1_name, (const unsigned char *)f1_data, strlen(f1_data))) {
        printf("failed to write %s\n", f1_name);
        zipClose(zfile, NULL);
        return false;
    }

    printf("adding file: %s\n", f2_name);
    if (!zipper_add_buf(zfile, f2_name, (const unsigned char *)f2_data, strlen(f2_data))) {
        printf("failed to write %s\n", f2_name);
        zipClose(zfile, NULL);
        return false;
    }

    printf("adding file: %s\n", f3_name);
    if (!zipper_add_buf(zfile, f3_name, (const unsigned char *)f3_data, strlen(f3_data))) {
        printf("failed to write %s\n", f3_name);
        zipClose(zfile, NULL);
        return false;
    }

    printf("adding dir: %s\n", d1_name);
    if (!zipper_add_dir(zfile, d1_name)) {
        printf("failed to write dir %s\n", d1_name);
        zipClose(zfile, NULL);
        return false;
    }

    printf("adding file: %s\n", f4_name);
    if (!zipper_add_buf(zfile, f4_name, (const unsigned char *)f4_data, strlen(f4_data))) {
        printf("failed to write %s\n", f4_name);
        zipClose(zfile, NULL);
        return false;
    }

    zipClose(zfile, NULL);
    return true;
}


size_t rhea::File::writeFile(const char *filename, const unsigned char *data, size_t len, bool append)
{
    FILE          *f;
    const char    *mode  = "w";
    size_t         wrote = 0;
    size_t         r;

    if (data == NULL){
    	return 0;
    }

    std::string myFile = "";

    if (filename == NULL || *filename == '\0'){
    	myFile = getFullName();
    } else {
    	myFile = filename;
    }

    if (myFile.empty()){
    	return 0;
    }

    if (append)
        mode = "a";

    f = fopen(myFile.c_str(), mode);
    if (f == NULL)
        return 0;

    do {
        r = fwrite(data, sizeof(*data), len, f);
        wrote += r;
        len   -= r;
        data  += r;
    } while (r > 0 && len != 0);

    fclose(f);
    return wrote;
}

rhea::Files* rhea::File::decompresZip(std::string dir, bool writeToDisk) {
    unzFile          uzfile;
    char            *zfilename;
    unsigned char   *buf;
    size_t           buflen;
    zipper_result_t  zipper_ret;
    uint64_t         len;

    rhea::Files* myFiles = new rhea::Files();
    std::vector<std::string> createdFiles = {};
    std::string cc = getFullName();
    const char *zipname = cc.c_str();
//    std::cout <<"Delam iz datoteke: " << zipname << std::endl;
    uzfile = unzOpen64(zipname);
    if (uzfile == NULL) {
        printf("Could not open %s for unzipping\n", zipname);
        return myFiles;
    }

    if (!dir.ends_with('/')){
    	dir += '/';
    }

    do {
        zipper_ret = ZIPPER_RESULT_SUCCESS;
//        std::cout << "deljam" << std::endl;
        zfilename  = zipper_filename(uzfile, NULL);
        if (zfilename == NULL){
        	std::cout << "filename v zip je v problem" << std::endl;
        	return myFiles;
        }

        if (zipper_isdir(uzfile)) {
//            std::cout<< "mam direktorij: " << zfilename << std::endl;
        	std::string agg = dir + zfilename;
            if (writeToDisk){
            	createPath(agg);
            }
            rhea::FileContent* dirdata = new rhea::FileContent;
            myFiles->add(agg,dirdata);
            createdFiles.push_back(agg);
            unzGoToNextFile(uzfile);
            free(zfilename);
            continue;
        }

        len = zipper_filesize(uzfile);
        zipper_ret = zipper_read_buf(uzfile, &buf, &buflen);
        if (zipper_ret == ZIPPER_RESULT_ERROR) {
            free(zfilename);
            break;
        }

        std::string some = dir + zfilename;
        std::size_t found = some.find_last_of('/');
        std::string aa = some.substr(0,found);
        if (writeToDisk){
        	createPath(aa);
        	writeFile(some.c_str(), buf, buflen, false);
        }
        rhea::FileContent* tmp = new rhea::FileContent;
//        std::cout << "dodajam v filecontent: " << zfilename << std::endl;
        tmp->data = (unsigned char *)malloc(sizeof(unsigned char)*buflen);
        std::copy(buf, buf+buflen,tmp->data);
//        std::cout << "dodal v filecontent: " << zfilename << std::endl;
        tmp->size =buflen;
        myFiles->add(some,tmp);
 //       createdFiles.push_back(some);
 //       std::cout << "dodajam v array: " << zfilename << std::endl;
        free(buf);
        free(zfilename);

    } while (zipper_ret == ZIPPER_RESULT_SUCCESS);

    if (zipper_ret == ZIPPER_RESULT_ERROR) {
        printf("failed to read file\n");
        return myFiles;
    }

    unzClose(uzfile);
    return myFiles;
}


unsigned char *rhea::File::readFile(const char *filename, size_t *len)
{
    FILE          *f;
    str_builder_t *sb;
    char          *out;
    char           temp[8192];
    size_t         mylen;
    size_t         r;

    if (len == NULL)
        len = &mylen;
    *len = 0;

    std::string myFile = "";

    if (filename == NULL) {
    	myFile = getFullName();
    } else {
    	myFile = filename;
    }

  	f = fopen(myFile.c_str(), "r");

   	if (f == NULL) {
   		std::cerr << "Error: " << strerror(errno);
        return NULL;
    }

    sb = str_builder_create();

    do {
        r = fread(temp, sizeof(*temp), sizeof(temp), f);
        if (r == 0){
        	break;
        }
        str_builder_add_str(sb, temp, r);
//        *len += r;
    } while (true);

    if (str_builder_len(sb) == 0) {
        fclose(f);
        str_builder_destroy(sb);
        return NULL;
    }
    fclose(f);

    out = str_builder_dump(sb, NULL);
    str_builder_destroy(sb);
    return (unsigned char *)out;
}

bool rhea::File::recurse_mkdir(const char *dirname)
{
    const char *p;
    char       *temp;
    bool        ret = true;

    temp = (char*) calloc(1, strlen(dirname)+1);
    /* Skip Windows drive letter. */
#ifdef _WIN32
    if ((p = strchr(dirname, ':') != NULL) {
        p++;
    } else {
#endif
        p = dirname;
#ifdef _WIN32
    }
#endif

    while ((p = strchr(p, SEP)) != NULL) {
        /* Skip empty elements. Could be a Windows UNC path or
           just multiple separators which is okay. */
        if (p != dirname && *(p-1) == SEP) {
            p++;
            continue;
        }
        /* Put the path up to this point into a temporary to
           pass to the make directory function. */
        memcpy(temp, dirname, p-dirname);
        temp[p-dirname] = '\0';
        p++;
#ifdef _WIN32
        if (CreateDirectory(temp, NULL) == FALSE) {
            if (GetLastError() != ERROR_ALREADY_EXISTS) {
                ret = false;
                break;
            }
        }
#else
        if (mkdir(temp, 0774) != 0) {
            if (errno != EEXIST) {
                ret = false;
                break;
            }
        }
#endif
    }
    free(temp);
    return ret;
}
