/*
 * SPDX-FileCopyrightText: 2022 Institute Rhea <eduard.filipas@rhea.si>
 *
 * SPDX-License-Identifier: EUPL-1.2
 */

#include <iostream>
#include "DbResultSet.h"

rhea::DbResultSet::DbResultSet(sqlite3 *db, const std::string& sql,std::vector<DbAtribute>* params) {
    sqlite3_stmt* stmt = {nullptr};
    sqlite3_prepare_v2(db,sql.c_str(),sql.length(),&stmt, nullptr);

    if (params) {
        for (int i = 0; i < params->size(); i++) {
            switch (params->at(i).getType()) {
                case 1: {
                    sqlite3_bind_int64(stmt, i+1, params->at(i).getAsLong());
                    break;
                }
                case 2: {
                    sqlite3_bind_double(stmt, i+1, params->at(i).getAsDouble());
                    break;
                }
                case 3: {
                    sqlite3_bind_text(stmt, i+1, params->at(i).getAsString().c_str(), params->at(i).getAsString().length(), SQLITE_TRANSIENT);
                    break;
                }
                case 4: {
                    sqlite3_bind_blob64(stmt, i+1, params->at(i).getAsBlob(), sizeof(params->at(i).getAsBlob()), SQLITE_TRANSIENT);
                    break;
                }
                default: {

                }
            }
        }
    }

    bool first = true;
    while ( sqlite3_step( stmt ) == SQLITE_ROW ) { // While query has result-rows.
        if (first) {
            currentRow = 0;
            for (int i = 0; i < sqlite3_column_count(stmt); i++) {
                DbAtributeMetaData* metaField = new DbAtributeMetaData();
                metaField->setName(sqlite3_column_name(stmt,i));
                metaField->setType(sqlite3_column_type(stmt, i));
                metaFields.push_back(*metaField);
            }
        }
        first = false;
        std::vector<DbAtribute*> columns = {};
        for (int i = 0; i < sqlite3_column_count(stmt); i++) {
            DbAtribute* field = new DbAtribute();
            field->setName(sqlite3_column_name(stmt,i));
            switch (sqlite3_column_type(stmt, i)) {
                case 1: {
                    field->setAsLong(sqlite3_column_int64(stmt, i));
                    break;
                }case 2:{
                    field->setAsDouble(sqlite3_column_double(stmt, i));
                    break;
                }case 3: {
                    field->setAsString(std::string(reinterpret_cast<const char *>(sqlite3_column_text(stmt, i))));
                    break;
                }case 4: {
                    unsigned char* tmp = reinterpret_cast<unsigned char *>(const_cast<void *>(sqlite3_column_blob(
                            stmt, i)));
                    field->setAsBlob(tmp);
                    break;
                }default:
                    field->setAsString("NULL");

            }
            columns.push_back(field);
        }
        results.push_back(columns);

    }

    sqlite3_finalize(stmt);

}

rhea::DbResultSet::~DbResultSet() {
    metaFields.clear();
    metaFields.shrink_to_fit();

    results.clear();
    results.shrink_to_fit();

}

int rhea::DbResultSet::rowCount() {
    return results.size();
}

int rhea::DbResultSet::columnCount() {
    return metaFields.size();
};

int rhea::DbResultSet::getRow(){
    return currentRow;
}

void rhea::DbResultSet::setRow(int val) {
    currentRow = val;
}

void rhea::DbResultSet::movePrevious() {
    currentRow = currentRow == -1 ? -1 : currentRow-1;
}


void rhea::DbResultSet::moveNext() {
    currentRow = currentRow == results.size() ? results.size() : currentRow +1;
}

void rhea::DbResultSet::moveFirst() {
    currentRow = 0;
}

void rhea::DbResultSet::moveLast() {
    currentRow = results.size()-1;
}

bool rhea::DbResultSet::isBof() const{
    return currentRow == -1;
}

bool rhea::DbResultSet::isEof() const{
    return currentRow == results.size();
}

rhea::DbAtribute* rhea::DbResultSet::getField(int row, int col){
    return results.at(row).at(col);
}

rhea::DbAtribute* rhea::DbResultSet::getField(int col){
    if (isBof() || isEof()){return nullptr;}
    return results.at(currentRow).at(col);
}

rhea::DbAtribute* rhea::DbResultSet::getField(std::string name){
    if (isBof() || isEof()){return nullptr;}

    int index = 0;
    bool found = false;
    for (auto atr : metaFields) {
        if(atr.getName()==name) {
            found= true;
            break;
        }
        index++;
    }

    if (found){
        return results.at(currentRow).at(index);
    }
    return nullptr;
}

std::string rhea::DbResultSet::getFieldName(int index) {
    return metaFields.at(index).getName();
}

int rhea::DbResultSet::getFieldType(int index) {
    return metaFields.at(index).getType();;
}

int rhea::DbResultSet::getFieldType(std::string name) {
    for (auto atr : metaFields) {
        if(atr.getName()==name) {
            return atr.getType();
        }
    }
    return -1;
}
